_integrity_exceptions = []
try:
    from asyncpg.exceptions import IntegrityConstraintViolationError

    _integrity_exceptions.append(IntegrityConstraintViolationError)
except ImportError:  # pragma: nocover
    "asyncpg not installed"


try:
    from aiosqlite import IntegrityError

    _integrity_exceptions.append(IntegrityError)
except ImportError:  # pragma: nocover
    "aiosqlite not installed"


INTEGRITY_CHECK_EXCEPTIONS = tuple(_integrity_exceptions)



"""
Table schema for tables in the license-manager backend
"""
import sqlalchemy
from sqlalchemy import Column, Integer, String, Table, func
from sqlalchemy.sql.schema import CheckConstraint, ForeignKey
from sqlalchemy.sql.sqltypes import DateTime
from sqlalchemy_utils import ScalarListType  # type: ignore

metadata = sqlalchemy.MetaData()


license_table = Table(
    "license",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("product_feature", String, unique=True),
    Column("used", Integer, CheckConstraint("used>=0")),
    Column("total", Integer, CheckConstraint("total>=0")),
    CheckConstraint("used<=total"),
)

license_searchable_fields = [
    license_table.c.product_feature,
]

license_sortable_fields = [
    license_table.c.product_feature,
    license_table.c.used,
    license_table.c.total,
]


booking_table = Table(
    "booking",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("job_id", String),
    Column("product_feature", String),
    Column("booked", Integer, CheckConstraint("booked>=0")),
    Column("lead_host", String),
    Column("user_name", String),
    Column("cluster_name", String),
    Column("created_at", DateTime, default=func.now()),
    Column("config_id", Integer, ForeignKey("config.id"), nullable=False),
)

booking_searchable_fields = [
    booking_table.c.job_id,
    booking_table.c.product_feature,
    booking_table.c.lead_host,
    booking_table.c.user_name,
    booking_table.c.cluster_name,
]

booking_sortable_fields = [
    booking_table.c.job_id,
    booking_table.c.product_feature,
    booking_table.c.booked,
    booking_table.c.lead_host,
    booking_table.c.user_name,
    booking_table.c.cluster_name,
]


config_table = Table(
    "config",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String, unique=True),
    Column("product", String),
    Column("features", String),
    Column("license_servers", ScalarListType(str)),
    Column("license_server_type", String),
    Column("grace_time", Integer, CheckConstraint("grace_time>=0")),
    Column("client_id", String, nullable=False, index=True),
)

config_searchable_fields = [
    config_table.c.name,
    config_table.c.product,
    config_table.c.features,
    config_table.c.license_server_type,
    config_table.c.client_id,
]

config_sortable_fields = [
    config_table.c.name,
    config_table.c.product,
    config_table.c.features,
    config_table.c.license_server_type,
    config_table.c.grace_time,
    config_table.c.client_id,
]



"""
Persistent data storage for the backend
"""
import typing

import databases
import sqlalchemy
from fastapi.exceptions import HTTPException
from sqlalchemy import Column, or_
from sqlalchemy.sql.expression import BooleanClauseList, UnaryExpression
from starlette import status

from lm_backend.config import settings
from lm_backend.table_schemas import metadata

database = databases.Database(settings.DATABASE_URL)


def create_all_tables():
    engine = sqlalchemy.create_engine(settings.DATABASE_URL)
    metadata.create_all(engine)


def render_sql(query) -> str:
    """
    Render a sqlalchemy query into a string for debugging.
    """
    return query.compile(dialect=database._backend._dialect, compile_kwargs={"literal_binds": True})


def search_clause(
    search_terms: str,
    searchable_fields: typing.List[Column],
) -> BooleanClauseList:
    """
    Create search clause across searchable fields with search terms.
    """
    return or_(*[field.ilike(f"%{term}%") for field in searchable_fields for term in search_terms.split()])


def sort_clause(
    sort_field: str,
    sortable_fields: typing.List[Column],
    sort_ascending: bool,
) -> typing.Union[Column, UnaryExpression]:
    """
    Create a sort clause given a sort field, the list of sortable fields, and a sort_ascending flag.
    """
    sort_field_names = [f.name for f in sortable_fields]
    try:
        index = sort_field_names.index(sort_field)
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Invalid sorting column requested: {sort_field}. Must be one of {sort_field_names}",
        )
    sort_column: typing.Union[Column, UnaryExpression] = sortable_fields[index]
    if not sort_ascending:
        sort_column = sort_column.desc()
    return sort_column



"""
API request and response schemas
"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field, validator


class ConfigurationRow(BaseModel):
    """
    A configuration row.
    """

    id: Optional[int] = Field(None)
    name: Optional[str] = Field(None)
    product: str
    features: str
    license_servers: List[str]
    license_server_type: str
    grace_time: int
    client_id: str

    class Config:
        orm_mode = True


class ConfigurationItem(BaseModel):
    """
    A configuration parsed item.
    """

    id: Optional[int] = Field(None)
    name: Optional[str] = Field(None)
    product: str
    features: dict
    license_servers: List[str]
    license_server_type: str
    grace_time: int
    client_id: str

    class Config:
        orm_mode = True


PRODUCT_FEATURE_RX = r"^.+?\..+$"


class BookingFeature(BaseModel):
    """
    One booked count for a single product.feature in a booking
    """

    product_feature: str = Field(..., regex=PRODUCT_FEATURE_RX)
    booked: int

    class Config:
        orm_mode = True


class Booking(BaseModel):
    """
    A booking for a jobid with a list of the features it requests
    """

    job_id: str
    features: List[BookingFeature]
    lead_host: str
    user_name: str
    cluster_name: str

    class Config:
        orm_mode = True


class BookingRow(BaseModel):
    """
    A flattened booking, suitable to be inserted into the database
    """

    id: Optional[int] = Field(None)
    job_id: str
    product_feature: str = Field(..., regex=PRODUCT_FEATURE_RX)
    booked: int
    config_id: int
    lead_host: str
    user_name: str
    cluster_name: str

    class Config:
        orm_mode = True


class BookingRowDetail(BookingRow):
    """
    A booking row with more detail
    """

    created_at: Optional[datetime]
    config_name: Optional[str]

    class Config:
        orm_mode = True


class LicenseUseBase(BaseModel):
    """
    Used/Total counts for a product.feature license category
    """

    product_feature: str = Field(..., regex=PRODUCT_FEATURE_RX)
    used: int


class LicenseUseReconcile(LicenseUseBase):
    """
    A reconcile [PATCH] Used/Total counts for a product.feature license category

    For creating items through the reconcile mechanism
    """

    total: int


class LicenseUseReconcileRequest(LicenseUseReconcile):
    """
    Used in the /reconcile request.
    """

    used_licenses: List


class LicenseUse(LicenseUseBase):
    """
    Used/Available/Total counts for a product.feature license category

    Returned by GET queries, including `available` for convenience
    """

    total: int

    available: Optional[int]

    @validator("available", always=True)
    def validate_available(cls, value, values):
        """
        Set available as a function of the other fields
        """
        return values["total"] - values["used"]


class LicenseUseWithBooking(LicenseUseBase):
    """
    Used/Available/Booked/Total counts for a product.feature license category.

    Returned by GET queries, including `available` for convenience.
    """

    total: int
    booked: Optional[int]
    available: Optional[int]

    @validator("available", always=True)
    def validate_available(cls, value, values):
        """
        Set available as a function of the other fields
        """
        return values["total"] - (values["used"] + values["booked"])


class LicenseUseBooking(LicenseUseBase):
    """
    A booking [PUT] object, specifying how many tokens are needed and no total
    """



"""
Booking objects and routes
"""
from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.sql import delete, join, select

from lm_backend.api.config import get_config_id_for_product_features
from lm_backend.api.permissions import Permissions
from lm_backend.api_schemas import Booking, BookingRow, BookingRowDetail, LicenseUse, LicenseUseBooking
from lm_backend.compat import INTEGRITY_CHECK_EXCEPTIONS
from lm_backend.security import guard
from lm_backend.storage import database, search_clause, sort_clause
from lm_backend.table_schemas import (
    booking_searchable_fields,
    booking_sortable_fields,
    booking_table,
    config_table,
    license_table,
)

router = APIRouter()


@router.get(
    "/all",
    response_model=List[BookingRow],
    dependencies=[Depends(guard.lockdown(Permissions.BOOKING_VIEW))],
)
async def get_bookings_all(
    cluster_name: Optional[str] = Query(None),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
):
    """
    All license counts we are tracking, with the possibility to filter by cluster_name.

    Note that search and sort are not applied if the cluster_name parameter is supplied.
    """
    query = booking_table.select()
    if cluster_name:
        query = query.where(booking_table.c.cluster_name == cluster_name)
    else:
        if search is not None:
            query = query.where(search_clause(search, booking_searchable_fields))
        if sort_field is not None:
            query = query.order_by(sort_clause(sort_field, booking_sortable_fields, sort_ascending))
    fetched = await database.fetch_all(query)
    return [BookingRow.parse_obj(x) for x in fetched]


@router.get(
    "/{booking_id}",
    response_model=BookingRowDetail,
    dependencies=[Depends(guard.lockdown(Permissions.BOOKING_VIEW))],
)
async def get_booking(booking_id: int):
    """
    Get details of a single booking by its id.

    Includes the config_name in the returned object.
    """
    query = (
        select([config_table.c.name.label("config_name"), *booking_table.c])
        .select_from(
            join(
                config_table,
                booking_table,
                booking_table.c.config_id == config_table.c.id,
            )
        )
        .where(booking_table.c.id == booking_id)
        .order_by(booking_table.c.product_feature)
    )
    fetched = await database.fetch_one(query)
    return BookingRowDetail.parse_obj(fetched)


@router.get(
    "/job/{job_id}",
    response_model=List[BookingRow],
    dependencies=[Depends(guard.lockdown(Permissions.BOOKING_VIEW))],
)
async def get_bookings_job(job_id: str):
    """
    All bookings of a particular job
    """
    query = (
        booking_table.select()
        .where(booking_table.c.job_id == job_id)
        .order_by(booking_table.c.product_feature)
    )
    fetched = await database.fetch_all(query)
    return [BookingRow.parse_obj(x) for x in fetched]


async def _is_booking_available(booking: Booking):
    """Check if the total available is greater than the sum of in_use_total and new_booked, if it is then
    there is no more bookings available. The in_use_total is calculated accounting for all bookings.
    """
    query = booking_table.select()
    fetched = await database.fetch_all(query)
    all_bookings = [BookingRow.parse_obj(x) for x in fetched]

    query = license_table.select()
    fetched = await database.fetch_all(query)
    all_licenses = [LicenseUse.parse_obj(x) for x in fetched]

    for feature in booking.features:
        in_use_total = 0
        total = 0
        for license in all_licenses:
            if feature.product_feature == license.product_feature:
                in_use_total = license.used
                total = license.total
                break
        for book in all_bookings:
            if feature.product_feature == book.product_feature:
                in_use_total += book.booked

        insert_quantity = feature.booked

        if insert_quantity + in_use_total > total:
            return False
    return True


@database.transaction()
@router.put(
    "/book",
    dependencies=[Depends(guard.lockdown(Permissions.BOOKING_EDIT))],
)
async def create_booking(booking: Booking):
    """
    Put a LicenseUse booking object in the database, reserving some tokens

    If an existing product_feature exists for this, update it by incrementing `booked',
    otherwise create it.

    An error occurs if the new total for `booked` exceeds `total`
    """
    if not await _is_booking_available(booking):
        raise HTTPException(
            status_code=400,
            detail=f"Couldn't book {booking.job_id}, not enough licenses available.",
        )
    # update the booking table
    inserts = []

    for feature in booking.features:
        inserts.append(
            BookingRow(
                job_id=booking.job_id,
                product_feature=feature.product_feature,
                booked=feature.booked,
                config_id=await get_config_id_for_product_features(feature.product_feature),
                lead_host=booking.lead_host,
                user_name=booking.user_name,
                cluster_name=booking.cluster_name,
            )
        )

    inserts_without_id = []
    for insert in inserts:
        inserts_without_id.append(insert.dict(exclude={"id"}))
    try:
        await database.execute_many(query=booking_table.insert(), values=[i for i in inserts_without_id])
    except INTEGRITY_CHECK_EXCEPTIONS as e:
        raise HTTPException(
            status_code=400,
            detail=f"Couldn't book {booking.job_id}, it is already booked\n{e}",
        )

    # update the license table
    lubs = []
    for feat in booking.features:
        lubs.append(
            LicenseUseBooking(
                product_feature=feat.product_feature,
                used=feat.booked,
            )
        )
    return dict(message=f"inserted {booking.job_id}")


@database.transaction()
@router.delete(
    "/book/{job_id}",
    dependencies=[Depends(guard.lockdown(Permissions.BOOKING_EDIT))],
)
async def delete_booking(job_id: str):
    """
    Deduct tokens from a LicenseUse booking in the database

    It is an error to use this method with a product_feature that isn't yet in the
    database.  Use reconcile or booking[PUT] to create the target first.

    An error occurs if the new total for `booked` is < 0
    """
    # check that we're deleting a booking that exists
    rows = await get_bookings_job(job_id)
    if not rows:
        raise HTTPException(
            status_code=400,
            detail=(f"Couldn't find booking {job_id} to delete"),
        )

    # update the booking table
    q = delete(booking_table).where(booking_table.c.job_id == job_id)
    await database.execute(q)

    # update the license table
    lubs = []
    for row in rows:
        lubs.append(
            LicenseUseBooking(
                product_feature=row.product_feature,
                used=-row.booked,
            )
        )

    return dict(
        message=f"deleted {job_id}",
    )



"""
Instantiates armada-security resources for auth on api endpoints using project settings.
Also provides a factory function for TokenSecurity to reduce boilerplate.
"""

import logging
from typing import List

from armasec import Armasec
from armasec.schemas import DomainConfig

from lm_backend.config import settings


def get_domain_configs() -> List[DomainConfig]:
    domain_configs = [
        DomainConfig(
            domain=settings.ARMASEC_DOMAIN,
            audience=settings.ARMASEC_AUDIENCE,
            debug_logger=logging.getLogger("armasec").debug if settings.ARMASEC_DEBUG else None,
        )
    ]
    if all(
        [
            settings.ARMASEC_ADMIN_DOMAIN,
            settings.ARMASEC_ADMIN_AUDIENCE,
            settings.ARMASEC_ADMIN_MATCH_KEY,
            settings.ARMASEC_ADMIN_MATCH_VALUE,
        ]
    ):
        domain_configs.append(
            DomainConfig(
                domain=settings.ARMASEC_ADMIN_DOMAIN,
                audience=settings.ARMASEC_ADMIN_AUDIENCE,
                match_keys={settings.ARMASEC_ADMIN_MATCH_KEY: settings.ARMASEC_ADMIN_MATCH_VALUE},
                debug_logger=logging.getLogger("armasec").debug if settings.ARMASEC_DEBUG else None,
            )
        )
    return domain_configs


guard = Armasec(domain_configs=get_domain_configs())



"""
Functions to help with field validations for sorting model instances.
"""
from typing import Optional

from fastapi import HTTPException, status

from lm_backend.api_schemas import LicenseUse, LicenseUseWithBooking


class LicenseUseWithBookingSortFieldChecker:
    """
    Validate if the sort field is a valid field in the model.
    """

    _model = LicenseUseWithBooking

    def __call__(self, sort_field: Optional[str] = None) -> Optional[str]:
        if sort_field is not None:
            if sort_field not in LicenseUseWithBooking.__fields__.keys():
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=(
                        f"Invalid sorting column requested: {sort_field}. "
                        f"Must be one of {self.available_fields()}."
                    ),
                )
        return sort_field

    @classmethod
    def available_fields(cls):
        return list(cls._model.__fields__.keys())


class LicenseUseSortFieldChecker:
    """
    Validate if the sort field is a valid field in the model.
    """

    _model = LicenseUse

    def __call__(self, sort_field: Optional[str] = None) -> Optional[str]:
        if sort_field is not None:
            if sort_field not in LicenseUse.__fields__.keys():
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=(
                        f"Invalid sorting column requested: {sort_field}. "
                        f"Must be one of {self.available_fields()}."
                    ),
                )
        return sort_field

    @classmethod
    def available_fields(cls):
        return list(cls._model.__fields__.keys())



from ast import literal_eval
from typing import Dict, List, Optional, Union

from armasec import TokenPayload
from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from lm_backend.api.permissions import Permissions
from lm_backend.api_schemas import ConfigurationItem, ConfigurationRow
from lm_backend.compat import INTEGRITY_CHECK_EXCEPTIONS
from lm_backend.security import guard
from lm_backend.storage import database, search_clause, sort_clause
from lm_backend.table_schemas import config_searchable_fields, config_sortable_fields, config_table

router = APIRouter()


@router.get(
    "/agent/all",
    response_model=List[ConfigurationItem],
)
async def get_all_configurations_by_client_id(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.CONFIG_VIEW)),
):
    """
    Query database for all configurations filtering by client_id.
    """
    # client_id identifies the cluster where the agent is running
    client_id = token_payload.client_id

    if not client_id:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=("Couldn't find a valid client_id in the access token."),
        )

    query = config_table.select().where(config_table.c.client_id == client_id)

    if search is not None:
        query = query.where(search_clause(search, config_searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, config_sortable_fields, sort_ascending))
    fetched = await database.fetch_all(query)
    config_rows = [ConfigurationRow.parse_obj(x) for x in fetched]
    return [
        ConfigurationItem(**item.dict(exclude={"features"}), features=literal_eval(item.features))
        for item in config_rows
    ]


@router.get(
    "/all",
    response_model=List[ConfigurationItem],
    dependencies=[Depends(guard.lockdown(Permissions.CONFIG_VIEW))],
)
async def get_all_configurations(
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
):
    """
    Query database for all configurations.
    """
    query = config_table.select()
    if search is not None:
        query = query.where(search_clause(search, config_searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, config_sortable_fields, sort_ascending))
    fetched = await database.fetch_all(query)
    config_rows = [ConfigurationRow.parse_obj(x) for x in fetched]
    return [
        ConfigurationItem(**item.dict(exclude={"features"}), features=literal_eval(item.features))
        for item in config_rows
    ]


@router.get(
    "/{config_id}",
    response_model=ConfigurationItem,
    dependencies=[Depends(guard.lockdown(Permissions.CONFIG_VIEW))],
)
async def get_configuration(config_id: int):
    """
    Get one configuration row based on a given id.
    """
    query = config_table.select().where(config_table.c.id == config_id)
    fetched = await database.fetch_one(query)
    if not fetched:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=(f"Couldn't get config {config_id}, " "that ID does not exist in the database"),
        )
    config_row = ConfigurationRow.parse_obj(fetched)
    return ConfigurationItem(
        **config_row.dict(exclude={"features"}), features=literal_eval(config_row.features)
    )


async def get_config_id_for_product_features(product_feature: str) -> Union[int, None]:
    """Return the config id for the given product_feature."""
    product, _ = product_feature.split(".")
    query = config_table.select().where(config_table.c.product == product)
    fetched = await database.fetch_one(query)
    config_row = ConfigurationRow.parse_obj(fetched)
    return config_row.id


@router.get("/", response_model=int)
async def get_config_id(product_feature: str):
    """Given the product_feature return the config_id."""
    _id = await get_config_id_for_product_features(product_feature)
    return _id


@database.transaction()
@router.post(
    "/",
    dependencies=[Depends(guard.lockdown(Permissions.CONFIG_EDIT))],
)
async def add_configuration(configuration: ConfigurationRow):
    """
    Add a configuration to the database for the first time.
    """
    query = config_table.insert().values(
        # It is necessary to exclude None so the database won't attempt to insert a null id
        **configuration.dict(exclude_none=True),
    )
    try:
        await database.execute(query)
    except INTEGRITY_CHECK_EXCEPTIONS:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=(f"Couldn't insert config {configuration.id}), it already exists."),
        )

    return dict(message=f"inserted {configuration.id}")


@database.transaction()
@router.put(
    "/{config_id}",
    dependencies=[Depends(guard.lockdown(Permissions.CONFIG_EDIT))],
)
async def update_configuration(
    config_id: int,
    product: Optional[str] = Body(None),
    features: Optional[str] = Body(None),
    license_servers: Optional[List[str]] = Body(None),
    license_server_type: Optional[str] = Body(None),
    grace_time: Optional[int] = Body(None),
    client_id: Optional[str] = Body(None),
):
    """
    Update a configuration row in the database with all of the
    optional arguments provided.
    """
    update_dict: Dict = {"id": config_id}
    if product is not None:
        update_dict["product"] = product
    if features is not None:
        update_dict["features"] = features
    if license_servers is not None:
        update_dict["license_servers"] = license_servers
    if license_server_type is not None:
        update_dict["license_server_type"] = license_server_type
    if grace_time is not None:
        update_dict["grace_time"] = grace_time
    if client_id is not None:
        update_dict["client_id"] = client_id
    q_update = config_table.update().where(config_table.c.id == config_id).values(update_dict)
    async with database.transaction():
        try:
            await database.execute(q_update)
        except INTEGRITY_CHECK_EXCEPTIONS as e:
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

    return dict(message=f"updated configuration id: {config_id}")


@database.transaction()
@router.delete(
    "/{config_id}",
    dependencies=[Depends(guard.lockdown(Permissions.CONFIG_EDIT))],
)
async def delete_configuration(config_id: int):
    """
    Delete a configuration from the database based on its id.
    """
    query = config_table.select().where(config_table.c.id == config_id)
    rows = await database.fetch_one(query)
    if not rows:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=(f"Couldn't find config id: {config_id} to delete, " "it does not exist in the database."),
        )
    q = config_table.delete().where(config_table.c.id == config_id)
    try:
        await database.execute(q)
    except Exception:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=(f"Couldn't delete config {config_id})"),
        )
    return dict(message=f"Deleted {config_id} from the configuration table.")



from enum import Enum


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting LM Backend routes.
    """

    BOOKING_VIEW = "license-manager:booking:view"
    BOOKING_EDIT = "license-manager:booking:edit"
    CONFIG_VIEW = "license-manager:config:view"
    CONFIG_EDIT = "license-manager:config:edit"
    LICENSE_VIEW = "license-manager:license:view"
    LICENSE_EDIT = "license-manager:license:edit"



import asyncio
from typing import Dict, List, Optional, Sequence, Tuple

from fastapi import APIRouter, Depends, Query
from sqlalchemy import func
from sqlalchemy.sql import join, select, update

from lm_backend.api.permissions import Permissions
from lm_backend.api_schemas import (
    BookingRow,
    LicenseUse,
    LicenseUseBase,
    LicenseUseReconcile,
    LicenseUseReconcileRequest,
    LicenseUseWithBooking,
)
from lm_backend.helpers import LicenseUseSortFieldChecker, LicenseUseWithBookingSortFieldChecker
from lm_backend.security import guard
from lm_backend.storage import database, search_clause
from lm_backend.table_schemas import booking_table, license_searchable_fields, license_table

PRODUCT_FEATURE_RX = r"^.+?\..+$"
router = APIRouter()


@router.get(
    "/all",
    response_model=List[LicenseUse],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_VIEW))],
)
async def licenses_all(
    search: Optional[str] = Query(None),
    sort_field: str = Depends(LicenseUseSortFieldChecker()),
    sort_ascending: bool = Query(True),
):
    """
    All license counts we are tracking
    """
    query = license_table.select().order_by(license_table.c.product_feature)

    if search is not None:
        query = query.where(search_clause(search, license_searchable_fields))

    fetched = await database.fetch_all(query)
    licenses = [LicenseUse.parse_obj(x) for x in fetched]

    if sort_field is not None:
        licenses = sorted(
            licenses, key=lambda license: getattr(license, sort_field), reverse=not sort_ascending
        )

    return licenses


@router.get(
    "/complete/all",
    response_model=List[LicenseUseWithBooking],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_VIEW))],
)
async def licenses_all_with_booking(
    search: Optional[str] = Query(None),
    sort_field: str = Depends(LicenseUseWithBookingSortFieldChecker()),
    sort_ascending: bool = Query(True),
):
    """
    All license counts we are tracking with booked tokens included.
    """
    query = (
        select([*license_table.c, func.sum(func.coalesce(booking_table.c.booked, 0)).label("booked")])
        .select_from(
            join(
                license_table,
                booking_table,
                license_table.c.product_feature == booking_table.c.product_feature,
                isouter=True,
            )
        )
        .group_by(
            license_table.c.id,
            license_table.c.product_feature,
        )
        .order_by(license_table.c.product_feature)
    )
    if search is not None:
        query = query.where(search_clause(search, license_searchable_fields))

    fetched = await database.fetch_all(query)
    licenses = [LicenseUseWithBooking.parse_obj(x) for x in fetched]

    if sort_field is not None:
        licenses = sorted(
            licenses, key=lambda license: getattr(license, sort_field), reverse=not sort_ascending
        )

    return licenses


@router.get(
    "/cluster_update",
    response_model=List[Dict],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_VIEW))],
)
async def licenses_and_bookings_to_update():
    """
    Get the actual value to update the cluster for each license.
    """
    query = license_table.select()
    fetched = await database.fetch_all(query)
    all_licenses = [LicenseUse.parse_obj(x) for x in fetched]

    query = booking_table.select()
    fetched = await database.fetch_all(query)
    all_bookings = [BookingRow.parse_obj(x) for x in fetched]

    licenses_to_update_data: List = []
    for license in all_licenses:
        licenses_to_update_data.append(
            {
                "product_feature": license.product_feature,
                "bookings_sum": sum(
                    [
                        booking.booked
                        for booking in all_bookings
                        if booking.product_feature == license.product_feature
                    ]
                ),
                "license_total": license.total,
                "license_used": license.used,
            }
        )
    return licenses_to_update_data


@router.get(
    "/use/{product}",
    response_model=List[LicenseUse],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_VIEW))],
)
async def licenses_product(product: str):
    """
    Used counts of all licenses, 1 product
    """
    query = (
        license_table.select()
        .where(license_table.c.product_feature.like(f"{product}.%"))
        .order_by(license_table.c.product_feature)
    )
    fetched = await database.fetch_all(query)
    return [LicenseUse.parse_obj(x) for x in fetched]


@router.get(
    "/use/{product}/{feature}",
    response_model=List[LicenseUse],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_VIEW))],
)
async def licenses_product_feature(product: str, feature: str):
    """
    Used counts of a product.feature category
    """
    query = license_table.select().where(license_table.c.product_feature == f"{product}.{feature}")
    fetched = await database.fetch_all(query)
    return [LicenseUse.parse_obj(x) for x in fetched]


async def _find_license_updates_and_inserts(
    licenses: Sequence[LicenseUseBase],
) -> Tuple[Dict, Dict]:
    """
    Return a dict of updates and a dict of inserts according to whether
    one of the LicenseUses is in the database, or being created
    """
    license_dict = {i.product_feature: i for i in licenses}

    q_updating = select([license_table.c.product_feature]).where(
        license_table.c.product_feature.in_(list(license_dict.keys()))
    )
    updating = [r[0] for r in await database.fetch_all(q_updating)]
    not_updating = list(set(license_dict.keys()) - set(updating))

    updates = {pf: license_dict[pf] for pf in updating}
    inserts = {pf: license_dict[pf] for pf in not_updating}

    return updates, inserts


async def _get_these_licenses(product_features: Sequence[str]) -> List[LicenseUse]:
    """
    Fetch the specific list of licenses matching the argument
    """
    fetch_query = (
        select([license_table])
        .where(license_table.c.product_feature.in_(product_features))
        .order_by(license_table.c.product_feature)
    )
    fetched = await database.fetch_all(fetch_query)
    return [LicenseUse.parse_obj(f) for f in fetched]


async def _delete_if_in_use_booking(license: LicenseUseReconcileRequest):
    """
    Check the database for the license, if it is booked, then delete it.
    """
    queries = []
    for used_license in license.used_licenses:
        query = (
            booking_table.select()
            .where(booking_table.c.lead_host == used_license["lead_host"])
            .where(booking_table.c.user_name == used_license["user_name"])
            .where(booking_table.c.booked == used_license["booked"])
            .where(booking_table.c.product_feature == license.product_feature)
        )
        queries.append(database.fetch_one(query))
    fetched = await asyncio.gather(*queries)
    bookings = [BookingRow.parse_obj(item) for item in fetched if item is not None]
    if not bookings:
        return

    delete_queries = []
    for booking in bookings:
        delete_query = booking_table.delete().where(booking_table.c.id == booking.id)
        delete_queries.append(database.execute(delete_query))
    await asyncio.gather(*delete_queries)


async def _clean_up_in_use_booking(
    reconcile_request: List[LicenseUseReconcileRequest],
) -> List[LicenseUseReconcile]:
    """
    For each license in the reconcile check if the {lead_host}{user_name}{booked} matches something in the
    database, then delete these bookings because they are already in use.
    """
    reconcile = []
    for license in reconcile_request:
        await _delete_if_in_use_booking(license)
        reconcile.append(LicenseUseReconcile(**license.dict(exclude={"used_licenses"})))

    return reconcile


@database.transaction()
@router.patch(
    "/reconcile",
    response_model=List[LicenseUse],
    dependencies=[Depends(guard.lockdown(Permissions.LICENSE_EDIT))],
)
async def reconcile_changes(reconcile_request: List[LicenseUseReconcileRequest]):
    """
    Set counts for models
    """
    reconcile = await _clean_up_in_use_booking(reconcile_request)
    updates, inserts = await _find_license_updates_and_inserts(reconcile)

    ops = []
    # update existing licenses
    for pf, license_use in updates.items():
        q_update = (
            update(license_table).where(license_table.c.product_feature == pf).values(**license_use.dict())
        )
        ops.append(database.execute(q_update))

    # insert new licenses
    ops.append(
        database.execute_many(query=license_table.insert(), values=[i.dict() for i in inserts.values()])
    )

    # wait for all updates and inserts at once
    await asyncio.gather(*ops)

    # query them back out to return them to the client
    requested_keys = [lu.product_feature for lu in reconcile]
    return await _get_these_licenses(requested_keys)



from fastapi import APIRouter

from lm_backend.api.booking import router as router_booking
from lm_backend.api.config import router as router_config
from lm_backend.api.license import router as router_license

api_v1 = APIRouter()
api_v1.include_router(router_license, prefix="/license", tags=["License"])
api_v1.include_router(router_booking, prefix="/booking", tags=["Booking"])
api_v1.include_router(router_config, prefix="/config", tags=["Config"])



""" License-manager backend, command line entrypoint

Run with e.g. `uvicorn lm_backend.main:app` OR
set `licensemanager2.backend.main.handler` as the ASGI handler
"""
import logging
import sys
from typing import cast

import sentry_sdk
from fastapi import FastAPI, Response, status
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

from lm_backend import storage
from lm_backend.api import api_v1
from lm_backend.config import settings

subapp = FastAPI(root_path=settings.ASGI_ROOT_PATH)
subapp.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

subapp.include_router(api_v1, prefix="/api/v1")

if settings.SENTRY_DSN:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        sample_rate=cast(float, settings.SENTRY_SAMPLE_RATE),  # The cast silences mypy
        environment=settings.DEPLOY_ENV,
    )
    subapp.add_middleware(SentryAsgiMiddleware)


@subapp.get(
    "/health",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={204: {"description": "API is healthy"}},
)
async def health_check():
    return Response(status_code=status.HTTP_204_NO_CONTENT)


app = FastAPI()
app.mount("/lm", subapp)


@app.on_event("startup")
def begin_logging():
    """
    Configure logging
    """
    logger.remove()
    logger.add(sys.stderr, level=settings.LOG_LEVEL)
    logger.info(f"Logging configured 📝 Level: {settings.LOG_LEVEL}")

    if settings.LOG_LEVEL_SQL:
        level_sql = getattr(logging, settings.LOG_LEVEL_SQL)

        engine_logger = logging.getLogger("sqlalchemy.engine")
        engine_logger.setLevel(level_sql)

        databases_logger = logging.getLogger("databases")
        databases_logger.setLevel(level_sql)

        logger.info(f"Database logging configured 📝 Level: {settings.LOG_LEVEL_SQL}")


@app.on_event("startup")
async def init_database():
    """
    Connect the database; create it if necessary
    """
    storage.create_all_tables()
    await storage.database.connect()
    logger.info("Database configured 💽")


@app.on_event("shutdown")
async def disconnect_database():
    """
    Disconnect the database
    """
    await storage.database.disconnect()



from enum import Enum
from typing import Optional

from pydantic import BaseSettings, Field, HttpUrl


class LogLevelEnum(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class Settings(BaseSettings):
    """
    App config.
    """

    DEPLOY_ENV: str = "LOCAL"

    # Sentry settings
    SENTRY_DSN: Optional[str] = None
    SENTRY_SAMPLE_RATE: Optional[float] = Field(1.0, gt=0.0, le=1.0)

    # vv should be specified as something like /staging
    # to match where the API is deployed in API Gateway
    ASGI_ROOT_PATH: str = ""

    # database to connect
    DATABASE_URL: str

    # log level (everything except sql tracing)
    LOG_LEVEL: LogLevelEnum = LogLevelEnum.INFO

    # log level (sql tracing)
    LOG_LEVEL_SQL: Optional[LogLevelEnum]

    # Security Settings. For details, see https://github.com/omnivector-solutions/armsec
    ARMASEC_DOMAIN: str
    ARMASEC_AUDIENCE: Optional[HttpUrl]
    ARMASEC_DEBUG: bool = Field(False)
    ARMASEC_ADMIN_DOMAIN: Optional[str]
    ARMASEC_ADMIN_AUDIENCE: Optional[HttpUrl]
    ARMASEC_ADMIN_MATCH_KEY: Optional[str]
    ARMASEC_ADMIN_MATCH_VALUE: Optional[str]

    class Config:
        env_file = ".env"


settings = Settings()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lm_backend', 'lm_backend.api']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils>=0.37.8,<0.38.0',
 'SQLAlchemy>=1.4.29,<2.0.0',
 'armasec>=0.11,<0.12',
 'asyncpg>=0.24.0,<0.25.0',
 'aws-psycopg2>=1.2.1,<2.0.0',
 'databases[postgresql]>=0.5.3,<0.6.0',
 'fastapi>=0.68.0,<0.69.0',
 'loguru>=0.5.3,<0.6.0',
 'python-dotenv>=0.19.0,<0.20.0',
 'sentry-sdk>=1.3.1,<2.0.0',
 'uvicorn>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'license-manager-backend',
    'version': '2.2.13',
    'description': 'Provides an API for managing license data',
    'long_description': '# License-manager Backend',
    'author': 'OmniVector Solutions',
    'author_email': 'info@omnivector.solutions',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/omnivector-solutions/license-manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
